#
# Requests SlideRule to process the provided region of interest
# and write the results to an HDF5 file.  Useful for large regions.
#

import sys
import time
import logging
import geopandas
import folium
import matplotlib.pyplot as plt
from sliderule import icesat2
from sliderule.stream2disk import Hdf5Writer, Hdf5Reader

def parse_command_line(args, cfg):
    i = 1
    while i < len(args):
        for entry in cfg:
            if args[i] == '--'+entry:
                if type(cfg[entry]) is str:
                    cfg[entry] = args[i + 1]
                elif type(cfg[entry]) is list:
                    l = []
                    while (i + 1) < len(args) and args[i + 1].isnumeric():
                        l.append(int(args[i + 1]))
                        i += 1
                    cfg[entry] = l
                elif type(cfg[entry]) is int:
                    cfg[entry] = int(args[i + 1])
                i += 1
        i += 1
    return cfg

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # positional command line parameters
    region_filename = sys.argv[1]

    # set script defaults
    cfg = {
        "run": 'write', # run, plot, list
        "url": 'icesat2sliderule.org',
        "asset": 'nsidc-s3',
        "filename": 'atl06.hdf5',
        "tolerance": 0.0,
        "cellsize": 0.01,
        "n_clusters": 1,
        "max_resources": 100000
    }

    # parse configuration parameters
    parse_command_line(sys.argv, cfg)

    # build region
    region = icesat2.toregion(region_filename, tolerance=cfg["tolerance"], cellsize=cfg["cellsize"], n_clusters=cfg["n_clusters"])

    # set processing parameter defaults
    parms = {
        "poly": region["poly"],
        "raster": region["raster"],
        "clusters": region["clusters"],
        "srt": icesat2.SRT_LAND,
        "cnf": icesat2.CNF_SURFACE_HIGH,
        "atl08_class": ["atl08_ground"],
        "ats": 10.0,
        "cnt": 10,
        "len": 40.0,
        "res": 20.0,
        "maxi": 1
    }

    # parse processing parametera
    parse_command_line(sys.argv, parms)

    # run script
    if cfg["run"] == 'write':

        # configure SlideRule
        icesat2.init(cfg["url"], False, cfg["max_resources"], loglevel=logging.INFO)

        # create hdf5 writer
        hdf5writer = Hdf5Writer(cfg["filename"], parms)

        # atl06 processing request
        perf_start = time.perf_counter()
        icesat2.atl06p(parms, cfg["asset"], callback=lambda resource, result, index, total : hdf5writer.run(resource, result, index, total))
        print("Completed in {:.3f} seconds of wall-clock time".format(time.perf_counter() - perf_start))

        # close hdf5 file
        hdf5writer.finish()

    elif cfg["run"] == 'read':

        # create hdf5 reader
        hdf5reader = Hdf5Reader(cfg["filename"])

        # display GeoDataFrame
        print(hdf5reader.gdf.describe())

    elif cfg["run"] == 'plot':

        # Create Plot
        fig = plt.figure(num=None, figsize=(24, 12))

        # Plot Encasing Polygon
        ax1 = plt.subplot(121)
        lons = [p["lon"] for p in region["poly"]]
        lats = [p["lat"] for p in region["poly"]]
        ax1.plot(lons, lats, linewidth=1.5, color='r', zorder=2)

        # Plot Cluster Polygons
        ax2 = plt.subplot(122)
        for cluster in region["clusters"]:
            lons = [p["lon"] for p in cluster]
            lats = [p["lat"] for p in cluster]
            ax2.plot(lons, lats, linewidth=1.5, color='r', zorder=2)

        # Show Plot
        plt.show()

    elif cfg["run"] == 'display':

        # read into pandas
        pregion = geopandas.read_file(region_filename)

        # build folium map
        m = folium.Map(location=[region["poly"][0]["lat"], region["poly"][0]["lon"]], zoom_start=5)
        folium.Choropleth(pregion, line_color='blue', line_weight=1.5).add_to(m)

        # draw containing polygon
        lons = [p["lon"] for p in region["poly"]]
        lats = [p["lat"] for p in region["poly"]]
        points = [(lat, lon) for lat,lon in zip(lats,lons)]
        folium.PolyLine(points, color="green", weight=2.0, opacity=1).add_to(m)

        # draw polygon clusters
        for cluster in region["clusters"]:
            lons = [p["lon"] for p in cluster]
            lats = [p["lat"] for p in cluster]
            points = [(lat, lon) for lat,lon in zip(lats,lons)]
            folium.PolyLine(points, color="red", weight=2.5, opacity=1).add_to(m)

        # display raster
        html_filename = "map.html"
        m.save(html_filename)

    elif cfg["run"] == 'list':

        # get polygon(s)
        polygon = region["poly"]
        if len(region["clusters"]) > 0:
            polygon = region["clusters"]

        # query CMR
        resources = icesat2.cmr(polygon=polygon)

        # display resources
        for resource in resources:
            print(resource)
        print("Total: ", len(resources))


# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import requests
import numpy
import json
import struct
import ctypes
import logging
import threading
from datetime import datetime, timedelta

###############################################################################
# GLOBALS
###############################################################################

service_url = None
server_lock = threading.Condition()
server_table = {}

max_errors_per_server = 3
max_pending_per_server = 3
max_retries_per_request = 5

verbose = False

request_timeout = (10, 60) # (connection, read) in seconds

logger = logging.getLogger(__name__)

recdef_table = {}

gps_epoch = datetime(1980, 1, 6)
tai_epoch = datetime(1970, 1, 1, 0, 0, 10)

eventformats = {
    "TEXT":     0,
    "JSON":     1
}

eventlogger = {
    0: logger.debug,
    1: logger.info,
    2: logger.warning,
    3: logger.error,
    4: logger.critical
}

handleexcept = {
    0: {"name": "ERROR",                    "fatal": True,  "expected": False },
    1: {"name": "TIMEOUT",                  "fatal": False, "expected": True },
    2: {"name": "RESOURCE_DOES_NOT_EXIST",  "fatal": True,  "expected": True },
    3: {"name": "EMPTY_SUBSET",             "fatal": True,  "expected": True }
}

datatypes = {
    "TEXT":     0,
    "REAL":     1,
    "INTEGER":  2,
    "DYNAMIC":  3
}

basictypes = {
    "INT8":     { "fmt": 'b', "size": 1, "nptype": numpy.int8 },
    "INT16":    { "fmt": 'h', "size": 2, "nptype": numpy.int16 },
    "INT32":    { "fmt": 'i', "size": 4, "nptype": numpy.int32 },
    "INT64":    { "fmt": 'q', "size": 8, "nptype": numpy.int64 },
    "UINT8":    { "fmt": 'B', "size": 1, "nptype": numpy.uint8 },
    "UINT16":   { "fmt": 'H', "size": 2, "nptype": numpy.uint16 },
    "UINT32":   { "fmt": 'I', "size": 4, "nptype": numpy.uint32 },
    "UINT64":   { "fmt": 'Q', "size": 8, "nptype": numpy.uint64 },
    "BITFIELD": { "fmt": 'x', "size": 0, "nptype": numpy.byte },    # unsupported
    "FLOAT":    { "fmt": 'f', "size": 4, "nptype": numpy.single },
    "DOUBLE":   { "fmt": 'd', "size": 8, "nptype": numpy.double },
    "TIME8":    { "fmt": 'Q', "size": 8, "nptype": numpy.byte },
    "STRING":   { "fmt": 's', "size": 1, "nptype": numpy.byte }
}

codedtype2str = {
    0:  "INT8",
    1:  "INT16",
    2:  "INT32",
    3:  "INT64",
    4:  "UINT8",
    5:  "UINT16",
    6:  "UINT32",
    7:  "UINT64",
    8:  "BITFIELD",
    9:  "FLOAT",
    10: "DOUBLE",
    11: "TIME8",
    12: "STRING"
}

###############################################################################
# CLASSES
###############################################################################

class TransientError(Exception):
    """Processing exception that can be retried"""
    pass

###############################################################################
# UTILITIES
###############################################################################

#
#  get the ip address of an available sliderule server
#
def __getserv(stream):
    global server_table
    with server_lock:
        try:
            server_available = False
            while not server_available:
                serv = min(server_table.items(), key=lambda x:x[1]["pending"])[0]
                if stream:
                    if server_table[serv]["pending"] < max_pending_per_server:
                        server_table[serv]["pending"] += 1
                        server_available = True
                    else:
                        if not server_lock.wait(1.0):
                            logger.debug("Timeout occurred waiting for thread to be notified")
                else:
                    server_available = True
        except:
            raise RuntimeError('No available urls')
    return serv

#
#  __errserv
#
def __errserv(serv, stream):
    global server_table, max_errors_per_server
    with server_lock:
        try:
            if stream:
                server_table[serv]["pending"] -= 1
            server_table[serv]["errors"] += 1
            logger.warning(serv + " encountered consecutive error " + str(server_table[serv]["errors"]))
            if server_table[serv]["errors"] > max_errors_per_server:
                logger.critical("Removing " + serv + " from list of available servers due to too many consecutive errors")
                server_table.pop(serv, None)
        except Exception as e:
            logger.debug(serv + " already removed from table")
        finally:
            if stream:
                server_lock.notify()

#
#  __clrserv
#
def __clrserv(serv, stream):
    global server_table
    with server_lock:
        try:
            if stream:
                server_table[serv]["pending"] -= 1
            server_table[serv]["errors"] = 0
        except Exception as e:
            pass
        finally:
            server_lock.notify()

#
#  __populate
#
def __populate(rectype):
    global recdef_table
    if rectype not in recdef_table:
        recdef_table[rectype] = source("definition", {"rectype" : rectype})
    return recdef_table[rectype]

#
#  __decode
#
def __decode(rectype, rawdata):
    """
    rectype: record type supplied in response (string)
    rawdata: payload supplied in response (byte array)
    """
    global recdef_table

    # initialize record
    rec = { "__rectype": rectype }

    # get/populate record definition #
    recdef = __populate(rectype)

    # iterate through each field in definition
    for fieldname in recdef.keys():

        # double underline (__) as prefix indicates meta data
        if fieldname.find("__") == 0:
            continue

        # get field properties
        field = recdef[fieldname]
        ftype = field["type"]
        offset = int(field["offset"] / 8)
        elems = field["elements"]
        flags = field["flags"]

        # do not process pointers
        if "PTR" in flags:
            continue

        # get endianess
        if "LE" in flags:
            endian = '<'
        else:
            endian = '>'

        # decode basic type
        if ftype in basictypes:

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / basictypes[ftype]["size"])

            # build format string
            fmt = endian + str(elems) + basictypes[ftype]["fmt"]

            # parse data
            value = struct.unpack_from(fmt, rawdata, offset)

            # set field
            if ftype == "STRING":
                rec[fieldname] = ctypes.create_string_buffer(value[0]).value.decode('ascii')
            elif is_array:
                rec[fieldname] = value
            else:
                rec[fieldname] = value[0]

        # decode user type
        else:

            # populate record definition (if needed) #
            subrecdef = __populate(ftype)

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / subrecdef["__datasize"])

            # return parsed data
            if is_array:
                rec[fieldname] = []
                for e in range(elems):
                    rec[fieldname].append(__decode(ftype, rawdata[offset:]))
                    offset += subrecdef["__datasize"]
            else:
                rec[fieldname] = __decode(ftype, rawdata[offset:])

    # return record #
    return rec

#
#  __parse
#
def __parse(stream, callbacks):
    """
    stream: request response stream
    """
    recs = []

    rec_size_size = 4
    rec_size_index = 0
    rec_size_rsps = []

    rec_size = 0
    rec_index = 0
    rec_rsps = []

    for line in stream.iter_content(None):

        i = 0
        while i < len(line):

            # Parse Record Size
            if(rec_size_index < rec_size_size):
                bytes_available = len(line)  - i
                bytes_remaining = rec_size_size - rec_size_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_size_rsps.append(line[i:i+bytes_to_append])
                rec_size_index += bytes_to_append
                if(rec_size_index >= rec_size_size):
                    raw = b''.join(rec_size_rsps)
                    rec_size = struct.unpack('<i', raw)[0]
                    rec_size_rsps.clear()
                i += bytes_to_append

            # Parse Record
            elif(rec_size > 0):
                bytes_available = len(line) - i
                bytes_remaining = rec_size - rec_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_rsps.append(line[i:i+bytes_to_append])
                rec_index += bytes_to_append
                if(rec_index >= rec_size):
                    # Decode Record
                    rawbits = b''.join(rec_rsps)
                    rectype = ctypes.create_string_buffer(rawbits).value.decode('ascii')
                    rawdata = rawbits[len(rectype) + 1:]
                    rec     = __decode(rectype, rawdata)
                    if callbacks != None and rectype in callbacks:
                        # Execute Call-Back on Record
                        callbacks[rectype](rec)
                    else:
                        # Append Record
                        recs.append(rec)
                    # Reset Record Parsing
                    rec_rsps.clear()
                    rec_size_index = 0
                    rec_size = 0
                    rec_index = 0
                i += bytes_to_append

    return recs

#
#  __logeventrec
#
def __logeventrec(rec):
    if verbose:
        eventlogger[rec['level']]('%s' % (rec["attr"]))

#
#  __raiseexceptrec
#
def __raiseexceptrec(rec):
    rc = rec["code"]
    if rc in handleexcept:
        if verbose:
            logger.info("%s exception <%d>: %s", handleexcept[rc]["name"], rc, rec["text"])
        if not handleexcept[rc]["expected"]:
            logger.critical("Unexpected error: %s", handleexcept[rc]["name"])
        if not handleexcept[rc]["fatal"]:
            raise TransientError()

###############################################################################
# APIs
###############################################################################

#
#  SOURCE
#
def source (api, parm={}, stream=False, callbacks={'eventrec': __logeventrec, 'exceptrec': __raiseexceptrec}):
    '''
    Perform API call to SlideRule service

    Parameters
    ----------
        api:        str
                    name of the SlideRule endpoint
        parm:       dict
                    dictionary of request parameters
        stream:     bool
                    whether the request is a **normal** service or a **stream** service (see `De-serialization <./SlideRule.html#de-serialization>`_ for more details)
        callbacks:  dict
                    record type callbacks (advanced use)

    Returns
    -------
    bytearray
        response data

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("icesat2sliderule.org")
        >>> rqst = {
        ...     "time": "NOW",
        ...     "input": "NOW",
        ...     "output": "GPS"
        ... }
        >>> rsps = sliderule.source("time", rqst)
        >>> print(rsps)
        {'time': 1300556199523.0, 'format': 'GPS'}
    '''
    rqst = json.dumps(parm)
    complete = False
    retries = max_retries_per_request
    rsps = []
    while (not complete) and (retries > 0):
        retries -= 1
        serv = __getserv(stream) # it throws a RuntimeError that must be caught by calling function
        try:
            url  = '%s/source/%s' % (serv, api)
            if not stream:
                rsps = requests.get(url, data=rqst, timeout=request_timeout).json()
            else:
                data = requests.post(url, data=rqst, timeout=request_timeout, stream=True)
                data.raise_for_status()
                rsps = __parse(data, callbacks)
            __clrserv(serv, stream)
            complete = True
        except requests.ConnectionError as e:
            logger.error("Failed to connect to endpoint {} ... retrying request".format(url))
            __errserv(serv, stream)
        except requests.Timeout as e:
            logger.error("Timed-out waiting for response from endpoint {} ... retrying request".format(url))
            __errserv(serv, stream)
        except requests.exceptions.ChunkedEncodingError as e:
            logger.error("Unexpected termination of response from endpoint {} ... retrying request".format(url))
            __errserv(serv, stream)
        except requests.HTTPError as e:
            if e.response.status_code == 503:
                logger.error("Server experiencing heavy load, stalling on request to {} ... will retry".format(url))
                __clrserv(serv, stream)
            else:
                logger.error("Invalid HTTP response from endpoint {} ... retrying request".format(url))
                __errserv(serv, stream)
        except TransientError as e:
            logger.warning("Recoverable error occurred at {} ... retrying request".format(url))
            __clrserv(serv, stream)
        except:
            __errserv(serv, stream)
            raise
    return rsps

#
#  SET_URL
#
def set_url (urls):
    '''
    Configure sliderule package with URL of service

    Parameters
    ----------
        urls:   str
                IP address or hostname of SlideRule service (note, there is a special case where the url is provided as a list of strings
                instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests
                to the exact set provided; this is used for testing and for local installations and can be ignored by most users)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("service.my-sliderule-server.org")
    '''
    global server_table, service_url
    with server_lock:
        service_url = None
        server_table = {}
        if type(urls) == list: # hardcoded list of sliderule server IP addresses
            for serv in urls:
                server_table["http://" + serv] = {"pending": 0, "errors": 0}
        elif type(urls) == str: # IP address of sliderule's service discovery
            service_url = "http://" + urls + "/discovery/"
        else:
            raise TypeError('expected ip address or hostname as a string or list of strings')
    # then update server table
    update_available_servers()

#
#  UPDATE_AVAIABLE_SERVERS
#
def update_available_servers ():
    '''
    Causes the SlideRule Python client to refresh the list of available processing nodes. This is useful when performing large processing
    requests where there is time for auto-scaling to change the number of nodes running.

    This function does nothing if the client has been initialized with a hardcoded list of servers.

    Returns
    -------
    int
        the number of available processing nodes

    Examples
    --------
        >>> import sliderule
        >>> sliderule.update_available_servers()
    '''
    global server_table, service_url
    with server_lock:
        if service_url != None:
            server_table = {}
            response = requests.get(service_url, data='{"service":"sliderule"}', timeout=request_timeout).json()
            for entry in response['members']:
                server_table["http://" + entry] = {"pending": 0, "errors": 0}
        num_servers = len(server_table)
        max_workers = num_servers * max_pending_per_server
    return (num_servers, max_workers)

#
#  SET_VERBOSE
#
def set_verbose (enable):
    '''
    Configure sliderule package for verbose logging

    Parameters
    ----------
        enable:     bool
                    whether or not user level log messages received from SlideRule generate a Python log message

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_verbose(True)

        The default behavior of Python log messages is for them to be displayed to standard output.
        If you want more control over the behavior of the log messages being display, create and configure a Python log handler as shown below:

        >>> # import packages
        >>> import logging
        >>> from sliderule import sliderule
        >>> # Configure Logging
        >>> sliderule_logger = logging.getLogger("sliderule.sliderule")
        >>> sliderule_logger.setLevel(logging.INFO)
        >>> # Create Console Output
        >>> ch = logging.StreamHandler()
        >>> ch.setLevel(logging.INFO)
        >>> sliderule_logger.addHandler(ch)
    '''
    global verbose
    verbose = (enable == True)

#
#  SET_MAX_ERRORS
#
def set_max_errors (max_errors):
    '''
    Configure sliderule package's maximum number of errors per node setting.  When the client makes a request to a processing node,
    if there is an error, it will retry the request to a different processing node (if available), but will keep the original processing
    node in the list of available nodes and increment the number of errors associated with it.  But if a processing node accumulates up
    to the **max_errors** number of errors, then the node is removed from the list of available nodes and will not be used in future
    processing requests.

    A call to ``update_available_servers`` or ``set_url`` is needed to restore a removed node to the list of available servers.

    Parameters
    ----------
        max_errors:     int
                        sets the maximum number of errors per node

    Raises
    ------
    TypeError
        max_errors must be a positive integer

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_max_errors(3)
    '''
    global max_errors_per_server
    if max_errors > 0:
        max_errors_per_server = max_errors
    else:
        raise TypeError('max errors must be greater than zero')

#
#  SET_MAX_PENDING
#
def set_max_pending (max_pending):
    '''
    Configure sliderule package's maximum number of pending requests per node setting.  When the client makes a request, it first finds the processing
    node with the fewest requests in progress (i.e. pending), and sends the request to that node.  If the node with the fewest pending requests has
    `max_pending` requests, it will wait until a processing node becomes available with less than `max_pending` requests.

    Parameters
    ----------
        max_pending:    int
                        sets the maximum number of pending requests per node

    Raises
    ------
    TypeError
        max_pending must be a positive integer

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_max_pending(1)
    '''
    global max_pending_per_server
    if max_pending > 0:
        max_pending_per_server = max_pending
    else:
        raise TypeError('max pending must be greater than zero')

#
# SET_REQUEST_TIMEOUT
#
def set_rqst_timeout (timeout):
    '''
    Sets the TCP/IP connection and reading timeouts for future requests made to sliderule servers.
    Setting it lower means the client will failover more quickly, but may generate false positives if a processing request stalls or takes a long time returning data.
    Setting it higher means the client will wait longer before designating it a failed request which in the presence of a persistent failure means it will take longer for the client to remove the node from its available servers list.

    Parameters
    ----------
        timeout:    tuple
                    (<connection timeout in seconds>, <read timeout in seconds>)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_rqst_timeout((10, 60))
    '''
    global request_timeout
    if type(timeout) == tuple:
        request_timeout = timeout
    else:
        raise TypeError('timeout must be a tuple (<connection timeout>, <read timeout>)')

#
# GPS2UTC
#
def gps2utc (gps_time, as_str=True, epoch=gps_epoch):
    '''
    Convert a GPS based time returned from SlideRule into a UTC time.

    Parameters
    ----------
        gps_time:   int
                    number of seconds since GPS epoch (January 6, 1980)
        as_str:     bool
                    if True, returns the time as a string; if False, returns the time as datatime object
        epoch:      datetime
                    the epoch used in the conversion, defaults to GPS epoch (Jan 6, 1980)

    Returns
    -------
    datetime
        UTC time (i.e. GMT, or Zulu time)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.gps2utc(1235331234)
        '2019-02-27 19:34:03'
    '''
    gps_time = epoch + timedelta(seconds=gps_time)
    tai_time = gps_time + timedelta(seconds=19)
    tai_timestamp = (tai_time - tai_epoch).total_seconds()
    utc_timestamp = datetime.utcfromtimestamp(tai_timestamp)
    if as_str:
        return str(utc_timestamp)
    else:
        return utc_timestamp

#
# GET DEFINITION
#
def get_definition (rectype, fieldname):
    '''
    Get the underlying format specification of a field in a return record.

    Parameters
    ----------
    rectype:    str
                the name of the type of the record (i.e. "atl03rec")
    fieldname:  str
                the name of the record field (i.e. "cycle")

    Returns
    -------
    dict
        description of each field; see the `sliderule.basictypes` variable for different field types

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("icesat2sliderule.org")
        >>> sliderule.get_definition("atl03rec", "cycle")
        {'fmt': 'H', 'size': 2, 'nptype': <class 'numpy.uint16'>}
    '''
    recdef = __populate(rectype)
    if fieldname in recdef and recdef[fieldname]["type"] in basictypes:
        return basictypes[recdef[fieldname]["type"]]
    else:
        return {}



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import copy
import json
import ssl
import urllib.request
import datetime
import logging
import concurrent.futures
import warnings
import numpy
import geopandas
import uuid
import base64
from osgeo import gdal
from osgeo import ogr
from osgeo import osr
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry import Polygon
from sklearn.cluster import KMeans
import sliderule
from sliderule import version

###############################################################################
# GLOBALS
###############################################################################

# create logger
logger = logging.getLogger(__name__)

# default asset
DEFAULT_ASSET="nsidc-s3"

# default standard data product version
DEFAULT_ICESAT2_SDP_VERSION='005'

# default maximum number of resources to process in one request
DEFAULT_MAX_REQUESTED_RESOURCES = 300
max_requested_resources = DEFAULT_MAX_REQUESTED_RESOURCES

# default maximum number of workers used for one request
DEFAULT_MAX_WORKERS_PER_NODE = 3

# icesat2 parameters
CNF_POSSIBLE_TEP = -2
CNF_NOT_CONSIDERED = -1
CNF_BACKGROUND = 0
CNF_WITHIN_10M = 1
CNF_SURFACE_LOW = 2
CNF_SURFACE_MEDIUM = 3
CNF_SURFACE_HIGH = 4
SRT_LAND = 0
SRT_OCEAN = 1
SRT_SEA_ICE = 2
SRT_LAND_ICE = 3
SRT_INLAND_WATER = 4
ALL_ROWS = -1
MAX_COORDS_IN_POLYGON = 16384
GT1L = 10
GT1R = 20
GT2L = 30
GT2R = 40
GT3L = 50
GT3R = 60
STRONG_SPOTS = (1, 3, 5)
WEAK_SPOTS = (2, 4, 6)
LEFT_PAIR = 0
RIGHT_PAIR = 1
SC_BACKWARD = 0
SC_FORWARD = 1

# gps-based epoch for delta times #
ATLAS_SDP_EPOCH = datetime.datetime(2018, 1, 1)

###############################################################################
# ICESAT2 UTILITIES
###############################################################################

#
# __calcspot
#
def __calcspot(sc_orient, track, pair):

    # spacecraft in forward orientation
    if sc_orient == SC_BACKWARD:
        if track == 1:
            if pair == LEFT_PAIR:
                return 1
            elif pair == RIGHT_PAIR:
                return 2
        elif track == 2:
            if pair == LEFT_PAIR:
                return 3
            elif pair == RIGHT_PAIR:
                return 4
        elif track == 3:
            if pair == LEFT_PAIR:
                return 5
            elif pair == RIGHT_PAIR:
                return 6

    # spacecraft in backward orientation
    elif sc_orient == SC_FORWARD:
        if track == 1:
            if pair == LEFT_PAIR:
                return 6
            elif pair == RIGHT_PAIR:
                return 5
        elif track == 2:
            if pair == LEFT_PAIR:
                return 4
            elif pair == RIGHT_PAIR:
                return 3
        elif track == 3:
            if pair == LEFT_PAIR:
                return 2
            elif pair == RIGHT_PAIR:
                return 1

    # unknown spot
    return 0

###############################################################################
# NSIDC UTILITIES
###############################################################################
# The functions below have been adapted from the NSIDC download script and
# carry the following notice:
#
# Copyright (c) 2020 Regents of the University of Colorado
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

CMR_URL = 'https://cmr.earthdata.nasa.gov'
CMR_PAGE_SIZE = 2000
CMR_FILE_URL = ('{0}/search/granules.json?provider=NSIDC_ECS'
                '&sort_key[]=start_date&sort_key[]=producer_granule_id'
                '&scroll=true&page_size={1}'.format(CMR_URL, CMR_PAGE_SIZE))

def __build_version_query_params(version):
    desired_pad_length = 3
    if len(version) > desired_pad_length:
        raise RuntimeError('Version string too long: "{0}"'.format(version))

    version = str(int(version))  # Strip off any leading zeros
    query_params = ''

    while len(version) <= desired_pad_length:
        padded_version = version.zfill(desired_pad_length)
        query_params += '&version={0}'.format(padded_version)
        desired_pad_length -= 1
    return query_params

def __cmr_filter_urls(search_results):
    """Select only the desired data files from CMR response."""
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return []

    entries = [e['links']
               for e in search_results['feed']['entry']
               if 'links' in e]
    # Flatten "entries" to a simple list of links
    links = list(itertools.chain(*entries))

    urls = []
    unique_filenames = set()
    for link in links:
        if 'href' not in link:
            # Exclude links with nothing to download
            continue
        if 'inherited' in link and link['inherited'] is True:
            # Why are we excluding these links?
            continue
        if 'rel' in link and 'data#' not in link['rel']:
            # Exclude links which are not classified by CMR as "data" or "metadata"
            continue

        if 'title' in link and 'opendap' in link['title'].lower():
            # Exclude OPeNDAP links--they are responsible for many duplicates
            # This is a hack; when the metadata is updated to properly identify
            # non-datapool links, we should be able to do this in a non-hack way
            continue

        filename = link['href'].split('/')[-1]
        if filename in unique_filenames:
            # Exclude links with duplicate filenames (they would overwrite)
            continue

        unique_filenames.add(filename)

        if ".h5" in link['href'][-3:]:
            resource = link['href'].split("/")[-1]
            urls.append(resource)

    return urls

def __cmr_granule_polygons(search_results):
    """Get the polygons for CMR returned granules"""
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return []
    granule_polygons = []
    # for each CMR entry
    for e in search_results['feed']['entry']:
        # for each polygon
        for polys in e['polygons']:
            coords = [float(i) for i in polys[0].split()]
            region = [{'lon':x,'lat':y} for y,x in zip(coords[::2],coords[1::2])]
            granule_polygons.append(region)
    # return granule polygons in sliderule region format
    return granule_polygons

def __cmr_search(short_name, version, time_start, time_end, **kwargs):
    """Perform a scrolling CMR query for files matching input criteria."""
    kwargs.setdefault('polygon',None)
    kwargs.setdefault('return_polygons',False)
    # build params
    params = '&short_name={0}'.format(short_name)
    params += __build_version_query_params(version)
    params += '&temporal[]={0},{1}'.format(time_start, time_end)
    if kwargs['polygon']:
        params += '&polygon={0}'.format(kwargs['polygon'])
    if kwargs['name_filter']:
        params += '&options[producer_granule_id][pattern]=true'
        params += '&producer_granule_id[]=' + kwargs['name_filter']
    cmr_query_url = CMR_FILE_URL + params
    logger.debug('cmr request={0}\n'.format(cmr_query_url))

    cmr_scroll_id = None
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    urls = []
    polys = []
    while True:
        req = urllib.request.Request(cmr_query_url)
        if cmr_scroll_id:
            req.add_header('cmr-scroll-id', cmr_scroll_id)
        response = urllib.request.urlopen(req, context=ctx)
        if not cmr_scroll_id:
            # Python 2 and 3 have different case for the http headers
            headers = {k.lower(): v for k, v in dict(response.info()).items()}
            cmr_scroll_id = headers['cmr-scroll-id']
            hits = int(headers['cmr-hits'])
        search_page = response.read()
        search_page = json.loads(search_page.decode('utf-8'))
        url_scroll_results = __cmr_filter_urls(search_page)
        if not url_scroll_results:
            break
        urls += url_scroll_results
        # append granule polygons
        if kwargs['return_polygons']:
            polygon_results = __cmr_granule_polygons(search_page)
        else:
            polygon_results = [None for _ in url_scroll_results]
        polys.extend(polygon_results)

    return (urls,polys)

###############################################################################
# SLIDERULE UTILITIES
###############################################################################

#
#  __get_values
#
def __get_values(data, dtype, size):
    """
    data:   tuple of bytes
    dtype:  element of codedtype
    size:   bytes in data
    """

    raw = bytes(data)
    datatype = sliderule.basictypes[sliderule.codedtype2str[dtype]]["nptype"]
    num_elements = int(size / numpy.dtype(datatype).itemsize)
    slicesize = num_elements * numpy.dtype(datatype).itemsize # truncates partial bytes
    values = numpy.frombuffer(raw[:slicesize], dtype=datatype, count=num_elements)

    return values

#
#  __query_resources
#
def __query_resources(parm, version, return_polygons=False):

    # Check Parameters are Valid
    if ("poly" not in parm) and ("t0" not in parm) and ("t1" not in parm):
        logger.error("Must supply some bounding parameters with request (poly, t0, t1)")
        return []

    # Submission Arguments for CRM #
    kwargs = {}
    kwargs['version'] = version
    kwargs['return_polygons'] = return_polygons

    # Pull Out Polygon #
    if "clusters" in parm and parm["clusters"] and len(parm["clusters"]) > 0:
        kwargs['polygon'] = parm["clusters"]
    elif "poly" in parm and parm["poly"] and len(parm["poly"]) > 0:
        kwargs['polygon'] = parm["poly"]

    # Pull Out Time Period #
    if "t0" in parm:
        kwargs['time_start'] = parm["t0"]
    if "t1" in parm:
        kwargs['time_end'] = parm["t1"]

    # Build Filters #
    name_filter_enabled = False
    rgt_filter = '????'
    if "rgt" in parm:
        rgt_filter = f'{parm["rgt"]}'.zfill(4)
        name_filter_enabled = True
    cycle_filter = '??'
    if "cycle" in parm:
        cycle_filter = f'{parm["cycle"]}'.zfill(2)
        name_filter_enabled = True
    region_filter = '??'
    if "region" in parm:
        region_filter = f'{parm["region"]}'.zfill(2)
        name_filter_enabled = True
    if name_filter_enabled:
        kwargs['name_filter'] = '*_' + rgt_filter + cycle_filter + region_filter + '_*'

    # Make CMR Request #
    if return_polygons:
        resources,polygons = cmr(**kwargs)
    else:
        resources = cmr(**kwargs)

    # Check Resources are Under Limit #
    if(len(resources) > max_requested_resources):
        logger.error("Exceeded maximum requested resources: %d (current max is %d)\nConsider using icesat2.set_max_resources to set a higher limit.", len(resources), max_requested_resources)
        resources = []
    else:
        logger.info("Identified %d resources to process", len(resources))

    # Return Resources #
    if return_polygons:
        return (resources,polygons)
    else:
        return resources

#
#  __emptyframe
#
def __emptyframe(**kwargs):
    # set default keyword arguments
    kwargs['crs'] = "EPSG:4326"
    return geopandas.GeoDataFrame(geometry=geopandas.points_from_xy([], []), crs=kwargs['crs'])

#
#  __todataframe
#
def __todataframe(columns, delta_time_key="delta_time", lon_key="lon", lat_key="lat", **kwargs):
    # set default keyword arguments
    kwargs['index_key'] = "time"
    kwargs['crs'] = "EPSG:4326"

    # Check Empty Columns
    if len(columns) <= 0:
        return __emptyframe(**kwargs)

    # Generate Time Column
    delta_time = (columns[delta_time_key]*1e9).astype('timedelta64[ns]')
    atlas_sdp_epoch = numpy.datetime64(ATLAS_SDP_EPOCH)
    columns['time'] = geopandas.pd.to_datetime(atlas_sdp_epoch + delta_time)

    # Generate Geometry Column
    geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key])
    del columns[lon_key]
    del columns[lat_key]

    # Create Pandas DataFrame object
    df = geopandas.pd.DataFrame(columns)

    # Build GeoDataFrame (default geometry is crs="EPSG:4326")
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])

    # Set index (default is Timestamp), can add `verify_integrity=True` to check for duplicates
    # Can do this during DataFrame creation, but this allows input argument for desired column
    gdf.set_index(kwargs['index_key'], inplace=True)

    # Sort values for reproducible output despite async processing
    gdf.sort_index(inplace=True)

    # Return GeoDataFrame
    return gdf

#
# __gdf2ploy
#
def __gdf2poly(gdf):
    # pull out coordinates
    hull = gdf.unary_union.convex_hull
    polygon = [{"lon": coord[0], "lat": coord[1]} for coord in list(hull.exterior.coords)]

    # determine winding of polygon #
    #              (x2               -    x1)             *    (y2               +    y1)
    wind = sum([(polygon[i+1]["lon"] - polygon[i]["lon"]) * (polygon[i+1]["lat"] + polygon[i]["lat"]) for i in range(len(polygon) - 1)])
    if wind > 0:
        # reverse direction (make counter-clockwise) #
        ccw_poly = []
        for i in range(len(polygon), 0, -1):
            ccw_poly.append(polygon[i - 1])
        # replace region with counter-clockwise version #
        polygon = ccw_poly

    # return polygon
    return polygon

#
#  __atl06
#
def __atl06 (parm, resource, asset):

    # Build ATL06 Request
    rqst = {
        "atl03-asset" : asset,
        "resource": resource,
        "parms": parm
    }

    # Execute ATL06 Algorithm
    rsps = sliderule.source("atl06", rqst, stream=True)

    # Flatten Responses
    columns = {}
    if len(rsps) <= 0:
        logger.debug("no response returned for %s", resource)
    elif (rsps[0]['__rectype'] != 'atl06rec' and rsps[0]['__rectype'] != 'atl06rec-compact'):
        logger.debug("invalid response returned for %s: %s", resource, rsps[0]['__rectype'])
    else:
        # Determine Record Type
        if rsps[0]['__rectype'] == 'atl06rec':
            rectype = 'atl06rec.elevation'
        else:
            rectype = 'atl06rec-compact.elevation'
        # Count Rows
        num_rows = 0
        for rsp in rsps:
            num_rows += len(rsp["elevation"])
        # Build Columns
        for field in rsps[0]["elevation"][0].keys():
            fielddef = sliderule.get_definition(rectype, field)
            if len(fielddef) > 0:
                columns[field] = numpy.empty(num_rows, fielddef["nptype"])
        # Populate Columns
        elev_cnt = 0
        for rsp in rsps:
            for elevation in rsp["elevation"]:
                for field in elevation.keys():
                    if field in columns:
                        columns[field][elev_cnt] = elevation[field]
                elev_cnt += 1

    # Return Response
    return __todataframe(columns, "delta_time", "lon", "lat"), resource


#
#  __atl03s
#
def __atl03s (parm, resource, asset):

    # Build ATL06 Request
    rqst = {
        "atl03-asset" : asset,
        "resource": resource,
        "parms": parm
    }

    # Execute ATL06 Algorithm
    rsps = sliderule.source("atl03s", rqst, stream=True)

    # Flatten Responses
    columns = {}
    if len(rsps) <= 0:
        logger.debug("no response returned for %s", resource)
    elif rsps[0]['__rectype'] != 'atl03rec':
        logger.debug("invalid response returned for %s: %s", resource, rsps[0]['__rectype'])
    else:
        # Count Rows
        num_rows = 0
        for rsp in rsps:
            num_rows += len(rsp["data"])
        # Build Columns
        for rsp in rsps:
            if len(rsp["data"]) > 0:
                # Allocate Columns
                for field in rsp.keys():
                    fielddef = sliderule.get_definition("atl03rec", field)
                    if len(fielddef) > 0:
                        columns[field] = numpy.empty(num_rows, fielddef["nptype"])
                for field in rsp["data"][0].keys():
                    fielddef = sliderule.get_definition("atl03rec.photons", field)
                    if len(fielddef) > 0:
                        columns[field] = numpy.empty(num_rows, fielddef["nptype"])
                break
        # Populate Columns
        ph_cnt = 0
        for rsp in rsps:
            ph_index = 0
            pair = 0
            left_cnt = rsp["count"][0]
            for photon in rsp["data"]:
                if ph_index >= left_cnt:
                    pair = 1
                for field in rsp.keys():
                    if field in columns:
                        if field == "count":
                            columns[field][ph_cnt] = pair
                        elif type(rsp[field]) is tuple:
                            columns[field][ph_cnt] = rsp[field][pair]
                        else:
                            columns[field][ph_cnt] = rsp[field]
                for field in photon.keys():
                    if field in columns:
                        columns[field][ph_cnt] = photon[field]
                ph_cnt += 1
                ph_index += 1
        # Rename Count Column to Pair Column
        columns["pair"] = columns.pop("count")

        # Create DataFrame
        df = __todataframe(columns, "delta_time", "longitude", "latitude")

        # Calculate Spot Column
        df['spot'] = df.apply(lambda row: __calcspot(row["sc_orient"], row["track"], row["pair"]), axis=1)

        # Return Response
        return df, resource

    # Error Case
    return __emptyframe(), resource

#
#  __parallelize
#
def __parallelize(callback, function, parm, resources, *args):

    # Update Available Servers #
    num_servers, max_workers = sliderule.update_available_servers()

    # Check if Servers are Available #
    if num_servers <= 0:
        logger.error("There are no servers available to fulfill this request")
        return __emptyframe()

    # Check Callback
    if callback == None:
        results = []

    # Make Parallel Processing Requests
    total_resources = len(resources)
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(function, parm, resource, *args) for resource in resources]

        # Wait for Results
        result_cnt = 0
        for future in concurrent.futures.as_completed(futures):
            result_cnt += 1
            result, resource = future.result()
            if len(result) > 0:
                if callback == None:
                    results.append(result)
                else:
                    callback(resource, result, result_cnt, total_resources)
            logger.info("%d points returned for %s (%d out of %d resources)", len(result), resource, result_cnt, total_resources)

    # Return Results
    if callback == None:
        if len(results) > 0:
            results.sort(key=lambda result: result.iloc[0]['delta_time'])
            return geopandas.pd.concat(results)
        else:
            return __emptyframe()

###############################################################################
# APIs
###############################################################################

#
#  INIT
#
def init (url, verbose=False, max_resources=DEFAULT_MAX_REQUESTED_RESOURCES, max_errors=3, loglevel=logging.CRITICAL):
    '''
    Initializes the underlying SlideRule module.  Must be called before other ICESat-2 API calls.
    This function is the same as calling the sliderule module functions: `set_url`, `set_verbose`, `set_max_errors`, along with the local `set_max_resources` function.

    Parameters
    ----------
        url :           str
                        the IP address or hostname of the SlideRule service (note, there is a special case where the url is provided as a list of strings instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests to the exact set provided; this is used for testing and for local installations and can be ignored by most users)
        verbose :       bool
                        whether or not user level log messages received from SlideRule generate a Python log message (see `sliderule.set_verbose <../user_guide/SlideRule.html#set_verbose>`_)
        max_errors :    int
                        the number of errors returned by a SlideRule server before the client drops it from the available server list
        max_resources : int
                        the maximum number of resources that are allowed to be processed in a single request
        loglevel :      int
                        minimum severity of log message to output

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init("my-sliderule-service.my-company.com", True)
    '''

    if verbose:
        loglevel = logging.INFO
    logging.basicConfig(level=loglevel)
    sliderule.set_url(url)
    sliderule.set_verbose(verbose)
    sliderule.set_max_errors(max_errors)
    set_max_resources(max_resources)

#
#  SET MAX RESOURCES
#
def set_max_resources (max_resources):
    '''
    Sets the maximum allowed number of resources to be processed in one request.  This is mainly provided as a sanity check for the user.

    Parameters
    ----------
        max_resources : int
                        the maximum number of resources that are allowed to be processed in a single request

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.set_max_resources(1000)
    '''
    global max_requested_resources
    max_requested_resources = max_resources

#
#  COMMON METADATA REPOSITORY
#
def cmr(**kwargs):
    '''
    Query the `NASA Common Metadata Repository (CMR) <https://cmr.earthdata.nasa.gov/search>`_ for a list of data within temporal and spatial parameters

    Parameters
    ----------
        polygon:    list
                    either a single list of longitude,latitude in counter-clockwise order with first and last point matching, defining region of interest (see `polygons <../user_guide/ICESat-2.html#polygons>`_), or a list of such lists when the region includes more than one polygon
        time_start: str
                    starting time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        time_end:   str
                    ending time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        version:    str
                    dataset version as found in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_
        short_name: str
                    dataset short name as defined in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_

    Returns
    -------
    list
        files (granules) for the dataset fitting the spatial and temporal parameters

    Examples
    --------
        >>> from sliderule import icesat2
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> granules = icesat2.cmr(polygon=region)
        >>> granules
        ['ATL03_20181017222812_02950102_003_01.h5', 'ATL03_20181110092841_06530106_003_01.h5', ... 'ATL03_20201111102237_07370902_003_01.h5']
    '''
    # set default polygon
    kwargs.setdefault('polygon', None)
    # set default start time to start of ICESat-2 mission
    kwargs.setdefault('time_start', '2018-10-13T00:00:00Z')
    # set default stop time to current time
    kwargs.setdefault('time_end', datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
    # set default version and product short name
    kwargs.setdefault('version', DEFAULT_ICESAT2_SDP_VERSION)
    kwargs.setdefault('short_name','ATL03')
    # return polygons for each requested granule
    kwargs.setdefault('return_polygons', False)
    # set default name filter
    kwargs.setdefault('name_filter', None)

    # return value
    resources = {} # [<url>] = <polygon>

    # create list of polygons
    polygons = [None]
    if kwargs['polygon'] and len(kwargs['polygon']) > 0:
        if type(kwargs['polygon'][0]) == dict:
            polygons = [copy.deepcopy(kwargs['polygon'])]
        elif type(kwargs['polygon'][0] == list):
            polygons = copy.deepcopy(kwargs['polygon'])

    # iterate through each polygon (or none if none supplied)
    for polygon in polygons:
        urls = []
        polys = []

        # issue CMR request
        for tolerance in [0.0001, 0.001, 0.01, 0.1, 1.0, None]:

            # convert polygon list into string
            polystr = None
            if polygon:
                flatpoly = []
                for p in polygon:
                    flatpoly.append(p["lon"])
                    flatpoly.append(p["lat"])
                polystr = str(flatpoly)[1:-1]
                polystr = polystr.replace(" ", "") # remove all spaces as this will be embedded in a url

            # call into NSIDC routines to make CMR request
            try:
                urls,polys = __cmr_search(kwargs['short_name'],
                                            kwargs['version'],
                                            kwargs['time_start'],
                                            kwargs['time_end'],
                                            polygon=polystr,
                                            return_polygons=kwargs['return_polygons'],
                                            name_filter=kwargs['name_filter'])
                break # exit loop because cmr search was successful
            except urllib.error.HTTPError as e:
                logger.error('HTTP Request Error: {}'.format(e.reason))
            except RuntimeError as e:
                logger.error("Runtime Error:", e)

            # simplify polygon
            if polygon and tolerance:
                raw_multi_polygon = [[(tuple([(c['lon'], c['lat']) for c in polygon]), [])]]
                shape = MultiPolygon(*raw_multi_polygon)
                buffered_shape = shape.buffer(tolerance)
                simplified_shape = buffered_shape.simplify(tolerance)
                simplified_coords = list(simplified_shape.exterior.coords)
                logger.warning('Using simplified polygon (for CMR request only!), {} points using tolerance of {}'.format(len(simplified_coords), tolerance))
                region = []
                for coord in simplified_coords:
                    point = {"lon": coord[0], "lat": coord[1]}
                    region.insert(0,point)
                polygon = region
            else:
                break # exit here because nothing can be done

        # populate resources
        for url, poly in zip(urls, polys):
            resources[url] = poly

    # build return lists
    url_list = list(resources.keys())
    poly_list = list(resources.values())

    if kwargs['return_polygons']:
        return (url_list,poly_list)
    else:
        return url_list

#
#  ATL06
#
def atl06 (parm, resource, asset=DEFAULT_ASSET):
    '''
    Performs ATL06-SR processing on ATL03 data and returns gridded elevations

    Parameters
    ----------
    parms:      dict
                parameters used to configure ATL06-SR algorithm processing (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
    resource:   str
                ATL03 HDF5 filename
    asset:      str
                data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    GeoDataFrame
        gridded elevations (see `Elevations <../user_guide/ICESat-2.html#elevations>`_)

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init("icesat2sliderule.org", True)
        >>> parms = { "cnf": 4, "ats": 20.0, "cnt": 10, "len": 40.0, "res": 20.0, "maxi": 1 }
        >>> resource = "ATL03_20181019065445_03150111_003_01.h5"
        >>> atl03_asset = "atlas-local"
        >>> rsps = icesat2.atl06(parms, resource, atl03_asset)
        >>> rsps
                dh_fit_dx  w_surface_window_final  ...                       time                     geometry
        0        0.000042               61.157661  ... 2018-10-19 06:54:46.104937  POINT (-63.82088 -79.00266)
        1        0.002019               61.157683  ... 2018-10-19 06:54:46.467038  POINT (-63.82591 -79.00247)
        2        0.001783               61.157678  ... 2018-10-19 06:54:46.107756  POINT (-63.82106 -79.00283)
        3        0.000969               61.157666  ... 2018-10-19 06:54:46.469867  POINT (-63.82610 -79.00264)
        4       -0.000801               61.157665  ... 2018-10-19 06:54:46.110574  POINT (-63.82124 -79.00301)
        ...           ...                     ...  ...                        ...                          ...
        622407  -0.000970               61.157666  ... 2018-10-19 07:00:29.606632  POINT (135.57522 -78.98983)
        622408   0.004620               61.157775  ... 2018-10-19 07:00:29.250312  POINT (135.57052 -78.98983)
        622409  -0.001366               61.157671  ... 2018-10-19 07:00:29.609435  POINT (135.57504 -78.98966)
        622410  -0.004041               61.157748  ... 2018-10-19 07:00:29.253123  POINT (135.57034 -78.98966)
        622411  -0.000482               61.157663  ... 2018-10-19 07:00:29.612238  POINT (135.57485 -78.98948)

        [622412 rows x 16 columns]
    '''
    try:
        return __atl06(parm, resource, asset)[0]
    except RuntimeError as e:
        logger.critical(e)
        return __emptyframe()

#
#  PARALLEL ATL06
#
def atl06p(parm, asset=DEFAULT_ASSET, version=DEFAULT_ICESAT2_SDP_VERSION, callback=None, resources=None):
    '''
    Performs ATL06-SR processing in parallel on ATL03 data and returns gridded elevations.  Unlike the `atl06 <#atl06>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL06-SR algorithm processing (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:          str
                        data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)
        version:        str
                        the version of the ATL03 data to use for processing
        callback:       bool
                        a callback function that is called for each resource and its results; when set, the API does not return anything (see `Callbacks <../user_guide/ICESat-2.html#callbacks>`_)

    Returns
    -------
    GeoDataFrame
        gridded elevations (see `Elevations <../user_guide/ICESat-2.html#elevations>`_)
    '''
    try:
        if resources == None:
            resources = __query_resources(parm, version)
        return __parallelize(callback, __atl06, parm, resources, asset)
    except RuntimeError as e:
        logger.critical(e)
        return __emptyframe()

#
#  Subsetted ATL03
#
def atl03s (parm, resource, asset=DEFAULT_ASSET):
    '''
    Subsets ATL03 data given the polygon and time range provided and returns segments of photons

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        resource:   str
                    ATL03 HDF5 filename
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    GeoDataFrame
        ATL03 extents (see `Photon Segments <../user_guide/ICESat-2.html#photon-segments>`_)
    '''
    try:
        return __atl03s(parm, resource, asset)[0]
    except RuntimeError as e:
        logger.critical(e)
        return __emptyframe()

#
#  PARALLEL SUBSETTED ATL03
#
def atl03sp(parm, asset=DEFAULT_ASSET, version=DEFAULT_ICESAT2_SDP_VERSION, callback=None, resources=None):
    '''
    Performs ATL03 subsetting in parallel on ATL03 data and returns photon segment data.  Unlike the `atl03s <#atl03s>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL03 subsetting (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:          str
                        data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)
        version:        str
                        the version of the ATL03 data to return
        callback:       bool
                        a callback function that is called for each resource and its results; when set, the API does not return anything (see `Callbacks <../user_guide/ICESat-2.html#callbacks>`_)

    Returns
    -------
    GeoDataFrame
        ATL03 segments (see `Photon Segments <../user_guide/ICESat-2.html#photon-segments>`_)
    '''
    try:
        if resources == None:
            resources = __query_resources(parm, version)
        return __parallelize(callback, __atl03s, parm, resources, asset)
    except RuntimeError as e:
        logger.critical(e)
        return __emptyframe()

#
#  H5
#
def h5 (dataset, resource, asset=DEFAULT_ASSET, datatype=sliderule.datatypes["DYNAMIC"], col=0, startrow=0, numrows=ALL_ROWS):
    '''
    Reads a dataset from an HDF5 file and returns the values of the dataset in a list

    This function provides an easy way for locally run scripts to get direct access to HDF5 data stored in a cloud environment.
    But it should be noted that this method is not the most efficient way to access remote H5 data, as the data is accessed one dataset at a time.
    Future versions may provide the ability to read multiple datasets at once, but in the meantime, if the user finds themselves needing direct
    access to a lot of HDF5 data residing in the cloud, then use of the H5Coro Python package is recommended as it provides a native Python package
    for performant direct access to HDF5 data in the cloud.

    One of the difficulties in reading HDF5 data directly from a Python script is converting format of the data as it is stored in the HDF5 to a data
    format that is easy to use in Python.  The compromise that this function takes is that it allows the user to supply the desired data type of the
    returned data via the **datatype** parameter, and the function will then return a **numpy** array of values with that data type.

    The data type is supplied as a ``sliderule.datatypes`` enumeration:

    - ``sliderule.datatypes["TEXT"]``: return the data as a string of unconverted bytes
    - ``sliderule.datatypes["INTEGER"]``: return the data as an array of integers
    - ``sliderule.datatypes["REAL"]``: return the data as an array of double precision floating point numbers
    - ``sliderule.datatypes["DYNAMIC"]``: return the data in the numpy data type that is the closest match to the data as it is stored in the HDF5 file

    Parameters
    ----------
        dataset:    str
                    full path to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``)
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset (see `Assets <#assets>`_)
        datatype:   int
                    the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        col:        int
                    the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        startrow:   int
                    the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        numrows:    int
                    the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Returns
    -------
    numpy array
        dataset values

    Examples
    --------
        >>> segments    = icesat2.h5("/gt1r/land_ice_segments/segment_id",  resource, asset)
        >>> heights     = icesat2.h5("/gt1r/land_ice_segments/h_li",        resource, asset)
        >>> latitudes   = icesat2.h5("/gt1r/land_ice_segments/latitude",    resource, asset)
        >>> longitudes  = icesat2.h5("/gt1r/land_ice_segments/longitude",   resource, asset)
        >>> df = pd.DataFrame(data=list(zip(heights, latitudes, longitudes)), index=segments, columns=["h_mean", "latitude", "longitude"])
    '''
    # Baseline Request
    rqst = {
        "asset" : asset,
        "resource": resource,
        "dataset": dataset,
        "datatype": datatype,
        "col": col,
        "startrow": startrow,
        "numrows": numrows,
        "id": 0
    }

    # Read H5 File
    try:
        rsps = sliderule.source("h5", rqst, stream=True)
    except RuntimeError as e:
        logger.critical(e)
        return numpy.empty(0)

    # Check if Data Returned
    if len(rsps) <= 0:
        return numpy.empty(0)

    # Build Record Data
    rsps_datatype = rsps[0]["datatype"]
    rsps_data = bytearray()
    rsps_size = 0
    for d in rsps:
        rsps_data += bytearray(d["data"])
        rsps_size += d["size"]

    # Get Values
    values = __get_values(rsps_data, rsps_datatype, rsps_size)

    # Return Response
    return values

#
#  H5P
#
def h5p (datasets, resource, asset=DEFAULT_ASSET):
    '''
    Reads a list of datasets from an HDF5 file and returns the values of the dataset in a dictionary of lists.

    This function is considerably faster than the ``icesat2.h5`` function in that it not only reads the datasets in
    parallel on the server side, but also shares a file context between the reads so that portions of the file that
    need to be read multiple times do not result in multiple requests to S3.

    For a full discussion of the data type conversion options, see `h5 <../user_guide/ICESat-2.html#h5>`_.

    Parameters
    ----------
        datasets:   dict
                    list of full paths to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``); see below for additional parameters that can be added to each dataset
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    dict
        numpy arrays of dataset values, where the keys are the dataset names

        The `datasets` dictionary can optionally contain the following elements per entry:

        * "valtype" (int): the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        * "col" (int): the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        * "startrow" (int): the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        * "numrows" (int): the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init(["127.0.0.1"], False)
        >>> datasets = [
        ...         {"dataset": "/gt1l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt1r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3r/land_ice_segments/h_li", "numrows": 5}
        ...     ]
        >>> rsps = icesat2.h5p(datasets, "ATL06_20181019065445_03150111_003_01.h5", "atlas-local")
        >>> print(rsps)
        {'/gt2r/land_ice_segments/h_li': array([45.3146427 , 45.27640582, 45.23608027, 45.21131015, 45.15692304]),
         '/gt2l/land_ice_segments/h_li': array([45.35118977, 45.33535027, 45.27195617, 45.21816889, 45.18534204]),
         '/gt1l/land_ice_segments/h_li': array([45.68811156, 45.71368944, 45.74234326, 45.74614113, 45.79866465]),
         '/gt3l/land_ice_segments/h_li': array([45.29602321, 45.34764226, 45.31430979, 45.31471701, 45.30034622]),
         '/gt1r/land_ice_segments/h_li': array([45.72632446, 45.76512574, 45.76337375, 45.77102473, 45.81307948]),
         '/gt3r/land_ice_segments/h_li': array([45.14954134, 45.18970635, 45.16637644, 45.15235916, 45.17135806])}
    '''
    # Baseline Request
    rqst = {
        "asset" : asset,
        "resource": resource,
        "datasets": datasets,
    }

    # Read H5 File
    try:
        rsps = sliderule.source("h5p", rqst, stream=True)
    except RuntimeError as e:
        logger.critical(e)
        rsps = []

    # Build Record Data
    results = {}
    for result in rsps:
        results[result["dataset"]] = __get_values(result["data"], result["datatype"], result["size"])

    # Return Results
    return results

#
# TO REGION
#
def toregion(source, tolerance=0.0, cellsize=0.01, n_clusters=1):
    '''
    Convert a GeoJSON representation of a set of geospatial regions into a list of lat,lon coordinates and raster image recognized by SlideRule

    Parameters
    ----------
        filename:   str
                    file name of GeoJSON formatted regions of interest, file **must** have named with the .geojson suffix
        tolerance:  float
                    tolerance used to simplify complex shapes so that the number of points is less than the limit (a tolerance of 0.001 typically works for most complex shapes)
        cellsize:   float
                    size of pixel in degrees used to create the raster image of the polygon
        clusters:   int
                    number of clusters of polygons to create when breaking up the request to CMR

    Returns
    -------
    dict
        a raster image and a list of longitudes and latitudes containing the region of interest that can be used for the **poly** and **raster** parameters in a processing request to SlideRule
        region = {
            "poly": [{"lat": <lat1>, "lon": <lon1>, ... }],
            "clusters": [{"lat": <lat1>, "lon": <lon1>, ... }, {"lat": <lat1>, "lon": <lon1>, ... }, ...],
            "raster": {
                "image": <base64 encoded geotiff image string>,
                "imagelength": <length of base64 encoded image>,
                "dimension": (<number of rows>, <number of columns>),
                "bbox": (<minimum longitutde>, <minimum latitude>, <maximum longitude>, <maximum latitude>),
                "cellsize": <cell size in degrees>
            }
        }

    Examples
    --------
        >>> from sliderule import icesat2
        >>> # Region of Interest #
        >>> region_filename = sys.argv[1]
        >>> region = icesat2.toregion(region_filename)
        >>> # Configure SlideRule #
        >>> icesat2.init("icesat2sliderule.org", False)
        >>> # Build ATL06 Request #
        >>> parms = {
        ...     "poly": region["poly"],
        ...     "srt": icesat2.SRT_LAND,
        ...     "cnf": icesat2.CNF_SURFACE_HIGH,
        ...     "ats": 10.0,
        ...     "cnt": 10,
        ...     "len": 40.0,
        ...     "res": 20.0,
        ...     "maxi": 1
        ... }
        >>> # Get ATL06 Elevations
        >>> atl06 = icesat2.atl06p(parms)
    '''
    # create:
    #   gdf - geodataframe
    #   inp_lyr - input layer
    if isinstance(source, geopandas.GeoDataFrame):
        # user provided GeoDataFrame instead of a file
        gdf = source

        # create input layer
        proj = osr.SpatialReference()
        proj.ImportFromEPSG(4326)
        rast_ogr_ds = ogr.GetDriverByName('Memory').CreateDataSource('wrk')
        inp_lyr = rast_ogr_ds.CreateLayer('poly', srs=proj)

        # add polygons to input layer
        for polygon in gdf.geometry:
            geom = ogr.CreateGeometryFromWkt(polygon.wkt)
            feat = ogr.Feature(inp_lyr.GetLayerDefn())
            feat.SetGeometryDirectly(geom)
            inp_lyr.CreateFeature(feat)

    elif isinstance(source, list) and (len(source) >= 4) and (len(source) % 2 == 0):
        # create lat/lon lists
        if len(source) == 4: # bounding box
            lons = [source[0], source[2], source[2], source[0], source[0]]
            lats = [source[1], source[1], source[3], source[3], source[1]]
        elif len(source) > 4: # polygon list
            lons = [source[i] for i in range(1,len(source),2)]
            lats = [source[i] for i in range(0,len(source),2)]

        # create geodataframe
        p = Polygon([point for point in zip(lons, lats)])
        gdf = geopandas.GeoDataFrame(geometry=[p], crs="EPSG:4326")

        # create input layer
        proj = osr.SpatialReference()
        proj.ImportFromEPSG(4326)
        rast_ogr_ds = ogr.GetDriverByName('Memory').CreateDataSource('wrk')
        inp_lyr = rast_ogr_ds.CreateLayer('poly', srs=proj)

        # add polygons to input layer
        for polygon in gdf.geometry:
            geom = ogr.CreateGeometryFromWkt(polygon.wkt)
            feat = ogr.Feature(inp_lyr.GetLayerDefn())
            feat.SetGeometryDirectly(geom)
            inp_lyr.CreateFeature(feat)

    elif isinstance(source, str) and ((source.find(".geojson") > 1) or (source.find(".shp") > 1)):
        # create geodataframe
        gdf = geopandas.read_file(source)

        # create input driver
        if (source.find(".geojson") > 1):
            inp_driver = ogr.GetDriverByName('GeoJSON')
        else: # (source.find(".shp") > 1)
            inp_driver = ogr.GetDriverByName('ESRI Shapefile')

        # create input layer
        inp_source = inp_driver.Open(source, 0)
        inp_lyr = inp_source.GetLayer()

    else:
        raise TypeError("incorrect filetype: please use a .geojson, .shp, or a geodataframe")

    # get extent of raster
    x_min, x_max, y_min, y_max = inp_lyr.GetExtent()
    x_ncells = int((x_max - x_min) / cellsize)
    y_ncells = int((y_max - y_min) / cellsize)

    # setup raster output
    out_driver = gdal.GetDriverByName('GTiff')
    out_filename = str(uuid.uuid4())
    out_source = out_driver.Create('/vsimem/' + out_filename, x_ncells, y_ncells, 1, gdal.GDT_Byte, options = [ 'COMPRESS=DEFLATE' ])
    out_source.SetGeoTransform((x_min, cellsize, 0, y_max, 0, -cellsize))
    out_source.SetProjection(inp_lyr.GetSpatialRef().ExportToWkt())
    out_lyr = out_source.GetRasterBand(1)
    out_lyr.SetNoDataValue(200)

    # rasterize
    gdal.RasterizeLayer(out_source, [1], inp_lyr, burn_values=[1])

    # close the data sources
    inp_source = None
    rast_ogr_ds = None
    out_source = None

    # read out raster data
    f = gdal.VSIFOpenL('/vsimem/' + out_filename, 'rb')
    gdal.VSIFSeekL(f, 0, 2)  # seek to end
    size = gdal.VSIFTellL(f)
    gdal.VSIFSeekL(f, 0, 0)  # seek to beginning
    raster = gdal.VSIFReadL(1, size, f)
    gdal.VSIFCloseL(f)

    # simplify polygon
    if(tolerance > 0.0):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            gdf = gdf.buffer(tolerance)
            gdf = gdf.simplify(tolerance)

    # generate polygon
    polygon = __gdf2poly(gdf)

    # generate clusters
    clusters = []
    if n_clusters > 1:
        # pull out centroids of each geometry object
        if "CenLon" in gdf and "CenLat" in gdf:
            X = numpy.column_stack((gdf["CenLon"],gdf["CenLat"]))
        else:
            s = gdf.centroid
            X = numpy.column_stack((s.x, s.y))
        # run k means clustering algorithm against polygons in gdf
        kmeans = KMeans(n_clusters=n_clusters, init='k-means++', random_state=5, max_iter=400)
        y_kmeans = kmeans.fit_predict(X)
        k = geopandas.pd.DataFrame(y_kmeans, columns=['cluster'])
        gdf = gdf.join(k)
        # build polygon for each cluster
        for n in range(n_clusters):
            c_gdf = gdf[gdf["cluster"] == n]
            c_poly = __gdf2poly(c_gdf)
            clusters.append(c_poly)

    # encode image in base64
    b64image = base64.b64encode(raster).decode('UTF-8')

    # return region #
    return {
        "poly": polygon, # convex hull of polygons
        "clusters": clusters, # list of polygon clusters for cmr request
        "raster": {
            "image": b64image, # geotiff image
            "imagelength": len(b64image), # encoded image size of geotiff
            "dimension": (y_ncells, x_ncells), # rows x cols
            "bbox": (x_min, y_min, x_max, y_max), # lon1, lat1 x lon2, lat2
            "cellsize": cellsize # in degrees
        }
    }

#
# GET VERSION
#
def get_version ():
    '''
    Get the version information for the running servers and Python client

    Returns
    -------
    dict
        dictionary of version information
    '''
    rsps = sliderule.source("version", {})
    rsps["client"] = {"version": version.full_version}
    return rsps



import sys
import json
import logging
import sliderule
from sliderule import icesat2
from sliderule import datatypes

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    url = ["127.0.0.1"]

    # Override server URL from command line
    if len(sys.argv) > 1:
        url = sys.argv[1]

    # Bypass service discovery
    if len(sys.argv) > 2:
        if sys.argv[2] == "bypass":
            url = [url]

    # Initialize ICESat2/SlideRule Package
    icesat2.init(url, False)

    # Query Version
    rsps = sliderule.source("version", {})

    # Display Version Information
    print(json.dumps(rsps, indent=4))


#
# Read ICESat-2 data using 'h5pyd' package and characterize read performance
#
# Prerequisites:    1. 'h5pyd' installed in local python environment
#                   2. HSDS server running and accessible to python client
#                   3. the hardcoded 'filepath' and 'filename' must be available to HSDS
#

import sys
import time
import h5pyd

timetable = {}
timetotal = 0

def printdata (file, dataset):
    data = file[dataset]
    print(data[0], data[1], data[2])

def timedata (file, dataset):
    global timetable, timetotal
    print("Reading " + dataset + ": ", end = '')
    start = time.time()
    datatotal = sum(file[dataset])
    end = time.time()
    duration = end - start
    timetable[dataset] = duration
    timetotal += duration
    print(duration, datatotal)

def timetrackdata (file, dataset):
    timedata(file, "/gt1l" + dataset)
    timedata(file, "/gt1r" + dataset)
    timedata(file, "/gt2l" + dataset)
    timedata(file, "/gt2r" + dataset)
    timedata(file, "/gt3l" + dataset)
    timedata(file, "/gt3r" + dataset)

if __name__ == '__main__':

    filepath = "/hsds/ATLAS/"
    filename = "ATL03_20181019065445_03150111_003_01.h5"
    test = "read"

    # Override test
    if len(sys.argv) > 1:
        test = sys.argv[1]

    # Override filename
    if len(sys.argv) > 2:
        filename = sys.argv[2]

    # Override filepath
    if len(sys.argv) > 3:
        filepath = sys.argv[3]

    # Open file
    print("Opening file: " + filepath + filename)
    h5file = h5pyd.File(filepath + filename, "r")

    # Test: Read
    if test == "read":
        printdata(h5file, "/gt1r/geolocation/segment_ph_cnt")
    elif test == "perf":
        timedata(h5file, "/ancillary_data/atlas_sdp_gps_epoch")
        timedata(h5file, "/orbit_info/sc_orient")
        timedata(h5file, "/ancillary_data/start_rgt")
        timedata(h5file, "/ancillary_data/end_rgt")
        timedata(h5file, "/ancillary_data/start_cycle")
        timedata(h5file, "/ancillary_data/end_cycle")
        timetrackdata(h5file, "/geolocation/delta_time")
        timetrackdata(h5file, "/geolocation/segment_ph_cnt")
        timetrackdata(h5file, "/geolocation/segment_id")
        timetrackdata(h5file, "/geolocation/segment_dist_x")
        timetrackdata(h5file, "/geolocation/reference_photon_lat")
        timetrackdata(h5file, "/geolocation/reference_photon_lon")
        timetrackdata(h5file, "/heights/dist_ph_along")
        timetrackdata(h5file, "/heights/h_ph")
        timetrackdata(h5file, "/heights/signal_conf_ph")
        timetrackdata(h5file, "/bckgrd_atlas/delta_time")
        timetrackdata(h5file, "/bckgrd_atlas/bckgrd_rate")
        print("TOTAL: " + str(timetotal))





#
# Query the consul server and display the available services
#

import sys
import requests
import json

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Set URL #
    ipaddr = "icesat2sliderule.org"
    if len(sys.argv) > 1:
        ipaddr = sys.argv[1]

    # Set Service #
    service = "sliderule"
    if len(sys.argv) > 2:
        service = sys.argv[2]

    # Query Services #
    response = requests.get("http://"+ipaddr+"/discovery/", data='{"service":"%s"}'%(service)).json()

    # Display Services #
    print(json.dumps(response, indent=2))
    print("Number of Nodes: {}".format(len(response['members'])))



#
# Imports
#
import sys
from sliderule import icesat2

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Set Filename
    filename = sys.argv[1]

    # Override Tolerance
    tolerance = 0.0
    if len(sys.argv) > 2:
        tolerance = float(sys.argv[2])

    # Override dataset
    dataset='ATL03'
    if len(sys.argv) > 3:
        dataset = sys.argv[3]

    # Override region of interest
    region = icesat2.toregion(filename, tolerance)

    # Query CMR for list of resources
    resources = icesat2.cmr(polygon=region["poly"], short_name=dataset)
    print("Region: {} points, {} files".format(len(region["poly"]), len(resources)))
    for resource in resources:
        print(resource)



import geopandas
import h5py
import datetime
import numpy
from sliderule import io
from sliderule import icesat2

class Hdf5Writer:

    def __init__(self, filename, parameters=None):
        self.filename = filename
        self.h5file = h5py.File(self.filename, mode='w')
        # add file attributes
        self.h5file.attrs['featureType'] = 'trajectory'
        self.h5file.attrs['title'] = "ATLAS/ICESat-2 SlideRule"
        self.h5file.attrs['reference'] = 'https://doi.org/10.5281/zenodo.5484048'
        self.h5file.attrs['date_created'] = datetime.datetime.now().isoformat()
        self.h5file.attrs['date_type'] = "UTC"
        self.h5file.attrs['time_type'] = "CCSDS UTC-A"
        # for each adjustable sliderule parameter
        for p in parameters:
            # try to get the parameter if available
            try:
                self.h5file.attrs[p] = parameters[p]
            except:
                # if empty or unavailable
                pass
        # save each region as a list attribute
        if "poly" in parameters:
            lon,lat = io.from_region(parameters["poly"])
            self.h5file.attrs['poly_x'] = lon.copy()
            self.h5file.attrs['poly_y'] = lat.copy()

    def run(self, resource, result, index, total):
        grp = self.h5file.create_group(resource[:-3])
        # convert geodataframe to pandas dataframe
        df = geopandas.pd.DataFrame(result.drop(columns='geometry'))
        df['latitude'] = result['geometry'].values.y
        df['longitude'] = result['geometry'].values.x
        # create dataset for each variable in the dataframe
        for key,val in df.items():
            grp.create_dataset(key, val.shape, data=val, dtype=val.dtype, compression='gzip')
        # set coordinate reference system as attribute
        if result.crs:
            grp.attrs['crs'] = str(result.crs)
        else:
            grp.attrs['crs'] = 'EPSG:4326'

    def finish(self):
        if self.h5file != None:
            self.h5file.close()
            self.h5file = None

    def __del__(self):
        self.finish()


class Hdf5Reader:

    def __init__(self, filename):
        self.filename = filename
        self.parameters = {}

        # open file
        h5file = h5py.File(self.filename, mode='r')

        # read in attributes
        for attr in h5file.attrs:
            self.parameters[attr] = h5file.attrs[attr]

        # read in datasets
        results = []
        for _,group in h5file.items():
            d = {}
            # populate columns
            for key,data in group.items():
                d[key] = data[:]
            # generate time column
            delta_time = (d['delta_time']*1e9).astype('timedelta64[ns]')
            atlas_sdp_epoch = numpy.datetime64(icesat2.ATLAS_SDP_EPOCH)
            d['time'] = geopandas.pd.to_datetime(atlas_sdp_epoch + delta_time)
            # generate geometry column
            geometry = geopandas.points_from_xy(d['longitude'],d['latitude'])
            # build DataFrame
            df = geopandas.pd.DataFrame(d)
            # build GeoDataFrame
            gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=group.attrs['crs'])
            # add to result list
            results.append(gdf)

        # build final GeoDataFrame
        self.gdf = geopandas.pd.concat(results)

        # close hdf5 file
        h5file.close()


#
# Imports
#
import sys
import json
import base64
import geopandas
import folium
from sliderule import icesat2
from sliderule import sliderule

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Make Call to TIFF Endpoint #
    filename = sys.argv[1]
    region = icesat2.toregion(filename)
    icesat2.init(["127.0.0.1"], False)
    rsps = sliderule.source("tiff", { "raster": region["raster"] })

    # Display Response #
    print(json.dumps(rsps, indent=4))

    # Optionally Write TIFF File #
    if "--write" in sys.argv[1:]:
        f = open("map.tiff", "wb")
        f.write(base64.b64decode(region["raster"]["image"]))
        f.close()

    # Generate Interactive Map #
    if "--interact" in sys.argv[1:]:
        pregion = geopandas.read_file(filename)
        m = folium.Map(location=[region["poly"][0]["lat"], region["poly"][0]["lon"]], zoom_start=5)
        folium.Choropleth(pregion, line_color='red', line_weight=2).add_to(m)
        m.save("map.html")

